@@ -2,7 +2,7 @@
  * xmpp.c       -- Jabber protocol handling
  *
  * Copyright (C) 2008-2014 Frank Zschockelt <mcabber@freakysoft.de>
- * Copyright (C) 2005-2014 Mikael Berthe <mikael@lilotux.net>
+ * Copyright (C) 2005-2015 Mikael Berthe <mikael@lilotux.net>
  * Parts come from the centericq project:
  * Copyright (C) 2002-2005 by Konstantin Klyagin <konst@konst.org.ua>
  *
@@ -17,9 +17,7 @@
  * General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
- * USA
+ * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 #include <stdlib.h>
 #include <string.h>
@@ -44,6 +42,10 @@
 
 #define RECONNECTION_TIMEOUT    60L
 
+#ifndef LOUDMOUTH_USES_SHA256
+#define FINGERPRINT_LENGTH      16  // old loudmouth still uses MD5 :(
+#endif
+
 LmConnection* lconnection = NULL;
 static guint AutoConnection;
 
@@ -297,10 +299,12 @@ void xmpp_request(const char *fjid, enum iqreq_type reqtype)
 }
 
 //  xmpp_send_msg(jid, text, type, subject,
-//                otrinject, *encrypted, type_overwrite)
+//                otrinject, *encrypted, type_overwrite, *xep184)
 // When encrypted is not NULL, the function set *encrypted to 1 if the
-// message has been PGP-encrypted.  If encryption enforcement is set and
-// encryption fails, *encrypted is set to -1.
+// message has been PGP (or OTR) -encrypted.  If encryption enforcement is set
+// and encryption fails, *encrypted is set to -1.
+// otrinject should be set to FALSE (unless the message already has an OTR
+// payload, i.e. if the function is called from an otr.c routine).
 void xmpp_send_msg(const char *fjid, const char *text, int type,
                    const char *subject, gboolean otrinject, gint *encrypted,
                    LmMessageSubType type_overwrite, gpointer *xep184)
@@ -309,6 +313,7 @@ void xmpp_send_msg(const char *fjid, const char *text, int type,
   LmMessageSubType subtype;
 #ifdef HAVE_LIBOTR
   int otr_msg = 0;
+  char *otr_msg_string = NULL;
 #endif
   char *barejid;
 #if defined HAVE_GPGME || defined XEP0085
@@ -352,13 +357,13 @@ void xmpp_send_msg(const char *fjid, const char *text, int type,
 #ifdef HAVE_LIBOTR
   if (otr_enabled() && !otrinject) {
     if (type == ROSTER_TYPE_USER) {
-      otr_msg = otr_send((char **)&text, barejid);
-      if (!text) {
-        g_free(barejid);
+      otr_msg_string = otr_send(text, barejid, &otr_msg);
+      if (!otr_msg_string) {
         if (encrypted)
           *encrypted = -1;
-        return;
+        goto xmpp_send_msg_return;
       }
+      text = otr_msg_string;
     }
     if (otr_msg && encrypted)
       *encrypted = ENCRYPTED_OTR;
@@ -382,13 +387,17 @@ void xmpp_send_msg(const char *fjid, const char *text, int type,
         key = settings_pgp_getkeyid(barejid);
         if (!key && res_pgpdata)
           key = res_pgpdata->sign_keyid;
-        if (key)
-          enc = gpg_encrypt(text, key);
+        if (key) {
+          int nkeys = 1;
+          const char *keys[] = { key, 0 };
+          if (carbons_enabled())
+            keys[nkeys++] = gpg_get_private_key_id();
+          enc = gpg_encrypt(text, keys, nkeys);
+        }
         if (!enc && force) {
           if (encrypted)
             *encrypted = -1;
-          g_free(barejid);
-          return;
+          goto xmpp_send_msg_return;
         }
       }
     }
@@ -413,10 +422,12 @@ void xmpp_send_msg(const char *fjid, const char *text, int type,
     g_free(enc);
   }
 
+#ifdef HAVE_LIBOTR
   // We probably don't want Carbons for encrypted messages, since the other
   // resources won't be able to decrypt them.
-  if (enc && carbons_enabled())
+  if (otr_msg && carbons_enabled())
     lm_message_node_add_child(x->node, "private", NS_CARBONS_2);
+#endif
 
   // XEP-0184: Message Receipts
   if (sl_buddy && xep184 &&
@@ -427,7 +438,6 @@ void xmpp_send_msg(const char *fjid, const char *text, int type,
                                   "xmlns", NS_RECEIPTS);
     *xep184 = g_strdup(lm_message_get_id(x));
   }
-  g_free(barejid);
 
 #ifdef XEP0085
   // If typing notifications are disabled, we can skip all this stuff...
@@ -461,6 +471,12 @@ xmpp_send_msg_no_chatstates:
     update_last_use();
   lm_connection_send(lconnection, x, NULL);
   lm_message_unref(x);
+
+xmpp_send_msg_return:
+#ifdef HAVE_LIBOTR
+  g_free(otr_msg_string);
+#endif
+  g_free(barejid);
 }
 
 #ifdef XEP0085
@@ -695,6 +711,9 @@ static LmSSLResponse ssl_cb(LmSSL *ssl, LmSSLStatus status, gpointer ud)
     break;
   case LM_SSL_STATUS_UNTRUSTED_CERT:
     scr_LogPrint(LPRINT_LOGNORM, "Certificate is not trusted!");
+    // The user specified a fingerprint, let's wait for lm to check that
+    if (settings_opt_get("ssl_fingerprint"))
+      return LM_SSL_RESPONSE_CONTINUE;
     break;
   case LM_SSL_STATUS_CERT_EXPIRED:
     scr_LogPrint(LPRINT_LOGNORM, "Certificate has expired!");
@@ -707,19 +726,24 @@ static LmSSLResponse ssl_cb(LmSSL *ssl, LmSSLStatus status, gpointer ud)
                  "Certificate hostname does not match expected hostname!");
     break;
   case LM_SSL_STATUS_CERT_FINGERPRINT_MISMATCH: {
-    char fpr[49];
-    fingerprint_to_hex((const unsigned char*)lm_ssl_get_fingerprint(ssl),
-                       fpr);
-    scr_LogPrint(LPRINT_LOGNORM,
-              "Certificate fingerprint does not match expected fingerprint!");
-    scr_LogPrint(LPRINT_LOGNORM, "Remote fingerprint: %s", fpr);
+#ifndef LOUDMOUTH_USES_SHA256
+      char fpr[3*FINGERPRINT_LENGTH] = {0};
+      fingerprint_to_hex(lm_ssl_get_fingerprint(ssl), fpr, FINGERPRINT_LENGTH);
+#endif
+      scr_LogPrint(LPRINT_LOGNORM,
+                "Certificate fingerprint does not match expected fingerprint!");
+#ifndef LOUDMOUTH_USES_SHA256
+      scr_LogPrint(LPRINT_LOGNORM, "Remote fingerprint: %s", fpr);
+#else
+      scr_LogPrint(LPRINT_LOGNORM, "Remote fingerprint: %s", lm_ssl_get_fingerprint(ssl));
+#endif
 
-    scr_LogPrint(LPRINT_LOGNORM, "Expected fingerprint: %s",
-                 settings_opt_get("ssl_fingerprint"));
+      scr_LogPrint(LPRINT_LOGNORM, "Expect fingerprint: %s",
+                   settings_opt_get("ssl_fingerprint"));
 
-    return LM_SSL_RESPONSE_STOP;
+      return LM_SSL_RESPONSE_STOP;
+    }
     break;
-  }
   case LM_SSL_STATUS_GENERIC_ERROR:
     scr_LogPrint(LPRINT_LOGNORM, "Generic SSL error!");
     break;
@@ -735,8 +759,20 @@ static LmSSLResponse ssl_cb(LmSSL *ssl, LmSSLStatus status, gpointer ud)
 static void connection_auth_cb(LmConnection *connection, gboolean success,
                                gpointer user_data)
 {
-  if (success) {
+  LmSSL *lssl;
+  if ((lssl = lm_connection_get_ssl(connection)) != NULL) {
+#ifndef LOUDMOUTH_USES_SHA256
+    char fpr[3*FINGERPRINT_LENGTH] = {0};
+    fingerprint_to_hex(lm_ssl_get_fingerprint(lssl), fpr, FINGERPRINT_LENGTH);
+    scr_LogPrint(LPRINT_LOGNORM, "Connection established.\n"
+                 "Remote fingerprint: %s", fpr);
+#else
+    scr_LogPrint(LPRINT_LOGNORM, "Connection established.\n"
+                 "Remote fingerprint: %s", lm_ssl_get_fingerprint(lssl));
+#endif
+  }
 
+  if (success) {
     xmpp_iq_request(NULL, NS_ROSTER);
     xmpp_iq_request(NULL, NS_DISCO_INFO);
     xmpp_request_storage("storage:bookmarks");
@@ -912,15 +948,16 @@ static void gotmessage(LmMessageSubType type, const char *from,
   if (rname) rname++;
 
 #ifdef HAVE_GPGME
-  if (enc && gpg_enabled()) {
-    decrypted_pgp = gpg_decrypt(enc);
-    if (decrypted_pgp) {
-      body = decrypted_pgp;
+  if (gpg_enabled()) {
+    if (enc) {
+      decrypted_pgp = gpg_decrypt(enc);
+      if (decrypted_pgp)
+        body = decrypted_pgp;
     }
+    // Check signature of the unencrypted/decrypted message
+    if (node_signed)
+      check_signature(bjid, rname, node_signed, body);
   }
-  // Check signature of an unencrypted message
-  if (node_signed && gpg_enabled())
-    check_signature(bjid, rname, node_signed, decrypted_pgp);
 #endif
 
   // Check for unexpected groupchat messages
@@ -1096,14 +1133,32 @@ static LmHandlerResult handle_messages(LmMessageHandler *handler,
   // Check for carbons!
   x = lm_message_node_find_xmlns(m->node, NS_CARBONS_2);
   gboolean carbons = FALSE;
-  if (x) {
+  if (x && (!g_strcmp0(x->name, "received") || !g_strcmp0(x->name, "sent"))) {
+    LmMessageNode *xenc;
+    const char *carbon_name = x->name;
     carbons = TRUE;
-    // Parse a message that is send to one of our other resources
-    if (!g_strcmp0(x->name, "received")) {
-      // Go 1 level deeper to the forwarded message
-      x = lm_message_node_find_xmlns(x, NS_FORWARD);
+    // Go 1 level deeper to the forwarded message
+    x = lm_message_node_find_xmlns(x, NS_FORWARD);
+    if (x)
       x = lm_message_node_get_child(x, "message");
 
+    if (!x) {
+      scr_LogPrint(LPRINT_LOGNORM,
+                   "Could not read carbon message!  Please fill a bug.");
+      goto handle_messages_return;
+    }
+
+    xenc = lm_message_node_find_xmlns(x, NS_ENCRYPTED);
+    if (xenc && (p = lm_message_node_get_value(xenc)) != NULL)
+      enc = p;
+
+    if (body && *body && !subject)
+      ns_signed = lm_message_node_find_xmlns(x, NS_SIGNED);
+    else
+      skip_process = TRUE;
+
+    // Parse a message that is send to one of our other resources
+    if (!g_strcmp0(carbon_name, "received")) {
       from = lm_message_node_get_attribute(x, "from");
       if (!from) {
         scr_LogPrint(LPRINT_LOGNORM, "Malformed carbon copy!");
@@ -1114,20 +1169,16 @@ static LmHandlerResult handle_messages(LmMessageHandler *handler,
       res = strchr(bjid, JID_RESOURCE_SEPARATOR);
       if (res) *res++ = 0;
 
-      if (body && *body && !subject && !enc)
-        ns_signed = lm_message_node_find_xmlns(x, NS_SIGNED);
-      else
-        skip_process = TRUE;
-
       // Try to handle forwarded chat state messages
       handle_state_events(from, res, x);
 
       scr_LogPrint(LPRINT_DEBUG, "Received incoming carbon from <%s>", from);
 
-    } else if (!g_strcmp0(x->name, "sent")) {
-      x = lm_message_node_find_xmlns(x, NS_FORWARD);
-      x = lm_message_node_get_child(x, "message");
-
+    } else if (!g_strcmp0(carbon_name, "sent")) {
+#ifdef HAVE_GPGME
+      char *decrypted_pgp = NULL;
+#endif
+      guint encrypted = 0;
       const char *to= lm_message_node_get_attribute(x, "to");
       if (!to) {
         scr_LogPrint(LPRINT_LOGNORM, "Malformed carbon copy!");
@@ -1136,10 +1187,32 @@ static LmHandlerResult handle_messages(LmMessageHandler *handler,
       g_free(bjid);
       bjid = jidtodisp(to);
 
+#ifdef HAVE_GPGME
+      if (gpg_enabled()) {
+        if (enc) {
+          decrypted_pgp = gpg_decrypt(enc);
+          if (decrypted_pgp) {
+            body = decrypted_pgp;
+            encrypted = ENCRYPTED_PGP;
+          }
+        }
+        /*
+        // Check messsage signature
+        // This won't work here, since check_signature wasn't intended
+        // to be used to check our own messages.
+        if (ns_signed)
+          check_signature(ME, NULL, ns_signed, body);
+        */
+      }
+#endif
+
       if (body && *body)
-        hk_message_out(bjid, NULL, timestamp, body, 0, NULL);
+        hk_message_out(bjid, NULL, timestamp, body, encrypted, TRUE, NULL);
 
       scr_LogPrint(LPRINT_DEBUG, "Received outgoing carbon for <%s>", to);
+#ifdef HAVE_GPGME
+      g_free(decrypted_pgp);
+#endif
       goto handle_messages_return;
     }
   } else { // Not a Carbon
@@ -1240,7 +1313,7 @@ static LmHandlerResult cb_caps(LmMessageHandler *h, LmConnection *c,
     LmMessageNode *info;
     LmMessageNode *query = lm_message_node_get_child(m->node, "query");
 
-    if (caps_has_hash(ver, bjid))
+    if (caps_has_hash(ver, bjid) || !query)
       goto caps_callback_return;
 
     caps_add(ver);
@@ -1495,6 +1568,7 @@ static LmHandlerResult handle_iq(LmMessageHandler *handler,
 {
   int i;
   const char *xmlns = NULL;
+  char *nodestr;
   LmMessageNode *x;
   LmMessageSubType mstype = lm_message_get_sub_type(m);
 
@@ -1505,8 +1579,9 @@ static LmHandlerResult handle_iq(LmMessageHandler *handler,
   }
 
   if (mstype == LM_MESSAGE_SUB_TYPE_RESULT) {
-    scr_LogPrint(LPRINT_DEBUG, "Unhandled IQ result? %s",
-                 lm_message_node_to_string(m->node));
+    nodestr = lm_message_node_to_string(m->node);
+    scr_LogPrint(LPRINT_DEBUG, "Unhandled IQ result? %s", nodestr);
+    g_free(nodestr);
     return LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS;
   }
 
@@ -1523,8 +1598,9 @@ static LmHandlerResult handle_iq(LmMessageHandler *handler,
       (mstype == LM_MESSAGE_SUB_TYPE_GET))
     send_iq_error(connection, m, XMPP_ERROR_NOT_IMPLEMENTED);
 
-  scr_LogPrint(LPRINT_DEBUG, "Unhandled IQ: %s",
-               lm_message_node_to_string(m->node));
+  nodestr = lm_message_node_to_string(m->node);
+  scr_LogPrint(LPRINT_DEBUG, "Unhandled IQ: %s", nodestr);
+  g_free(nodestr);
 
   scr_LogPrint(LPRINT_NORMAL, "Received unhandled IQ request from <%s>.",
                lm_message_get_from(m));
@@ -1695,14 +1771,16 @@ gint xmpp_connect(void)
 {
   const char *userjid, *password, *resource, *servername, *ssl_fpr;
   char *dynresource = NULL;
-  char fpr[16];
+#ifndef LOUDMOUTH_USES_SHA256
+  char fpr[FINGERPRINT_LENGTH] = {0};
+#endif
   const char *proxy_host;
   const char *resource_prefix = PACKAGE_NAME;
   char *fjid;
   int ssl, tls;
   LmSSL *lssl;
   unsigned int port;
-  unsigned int ping;
+  unsigned int ping = 40;
   LmMessageHandler *handler;
   GError *error = NULL;
 
@@ -1728,7 +1806,6 @@ gint xmpp_connect(void)
 
   g_log_set_handler("LM", LM_LOG_LEVEL_ALL, lm_debug_handler, NULL);
 
-  ping = 40;
   if (settings_opt_get("pinginterval"))
     ping = (unsigned int) settings_opt_get_int("pinginterval");
   lm_connection_set_keep_alive_rate(lconnection, ping);
@@ -1841,13 +1918,17 @@ gint xmpp_connect(void)
     port = (ssl ? LM_CONNECTION_DEFAULT_PORT_SSL : LM_CONNECTION_DEFAULT_PORT);
   lm_connection_set_port(lconnection, port);
 
-  if (ssl_fpr && (!hex_to_fingerprint(ssl_fpr, fpr))) {
+#ifndef LOUDMOUTH_USES_SHA256
+  if (ssl_fpr && (!hex_to_fingerprint(ssl_fpr, fpr, FINGERPRINT_LENGTH))) {
     scr_LogPrint(LPRINT_LOGNORM, "** Please set the fingerprint in the format "
                  "97:5C:00:3F:1D:77:45:25:E2:C5:70:EC:83:C8:87:EE");
     return -1;
   }
 
   lssl = lm_ssl_new((ssl_fpr ? fpr : NULL), ssl_cb, NULL, NULL);
+#else
+  lssl = lm_ssl_new(ssl_fpr, ssl_cb, NULL, NULL);
+#endif
   if (lssl) {
 #ifdef HAVE_LM_SSL_CIPHER_LIST
     const char *ssl_ciphers = settings_opt_get("ssl_ciphers");
@@ -1855,7 +1936,10 @@ gint xmpp_connect(void)
 #endif
 #ifdef HAVE_LM_SSL_CA
     const char *ssl_ca = settings_opt_get("ssl_ca");
-    lm_ssl_set_ca(lssl, ssl_ca);
+    char *ssl_ca_xp;
+    ssl_ca_xp = expand_filename(ssl_ca);
+    lm_ssl_set_ca(lssl, ssl_ca_xp);
+    g_free(ssl_ca_xp);
 #endif
     lm_ssl_use_starttls(lssl, !ssl, tls);
     lm_connection_set_ssl(lconnection, lssl);
@@ -1880,6 +1964,8 @@ void xmpp_insert_entity_capabilities(LmMessageNode *x, enum imstatus status)
 {
   LmMessageNode *y;
   const char *ver = entity_version(status);
+  if (!ver)
+    return;
 
   y = lm_message_node_add_child(x, "c", NULL);
   lm_message_node_set_attribute(y, "xmlns", NS_CAPS);
