@@ -2,7 +2,7 @@
  * xmpp_iqrequest.c -- Jabber IQ request handling
  *
  * Copyright (C) 2008-2010 Frank Zschockelt <mcabber@freakysoft.de>
- * Copyright (C) 2005-2010 Mikael Berthe <mikael@lilotux.net>
+ * Copyright (C) 2005-2014 Mikael Berthe <mikael@lilotux.net>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -15,9 +15,7 @@
  * General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
- * USA
+ * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
 #include <string.h>
@@ -31,6 +29,7 @@
 #include "settings.h"
 #include "hooks.h"
 #include "hbuf.h"
+#include "carbons.h"
 
 extern LmMessageNode *bookmarks;
 extern LmMessageNode *rosternotes;
@@ -47,6 +46,9 @@ static LmHandlerResult cb_ping(LmMessageHandler *h, LmConnection *c,
                                LmMessage *m, gpointer user_data);
 static LmHandlerResult cb_vcard(LmMessageHandler *h, LmConnection *c,
                                LmMessage *m, gpointer user_data);
+static LmHandlerResult cb_disco_info(LmMessageHandler *h, LmConnection *c,
+                               LmMessage *m, gpointer user_data);
+
 
 static struct IqRequestHandlers
 {
@@ -54,12 +56,13 @@ static struct IqRequestHandlers
   const gchar *querytag;
   LmHandleMessageFunction handler;
 } iq_request_handlers[] = {
-  {NS_ROSTER, "query", &cb_roster},
-  {NS_VERSION,"query", &cb_version},
-  {NS_TIME,   "query", &cb_time},
-  {NS_LAST,   "query", &cb_last},
-  {NS_PING,   "ping",  &cb_ping},
-  {NS_VCARD,  "vCard", &cb_vcard},
+  {NS_ROSTER,   "query", &cb_roster},
+  {NS_VERSION,  "query", &cb_version},
+  {NS_XMPP_TIME,"time",  &cb_time},
+  {NS_LAST,     "query", &cb_last},
+  {NS_PING,     "ping",  &cb_ping},
+  {NS_VCARD,    "vCard", &cb_vcard},
+  {NS_DISCO_INFO, "query", &cb_disco_info},
   {NULL, NULL, NULL}
 };
 
@@ -75,6 +78,26 @@ enum vcard_attr {
   vcard_pref    = 1<<7,
 };
 
+static LmHandlerResult cb_disco_info(LmMessageHandler *h, LmConnection *c,
+                               LmMessage *m, gpointer user_data)
+{
+  LmMessageNode *ansqry;
+  LmMessageNode *feature;
+
+  ansqry = lm_message_node_get_child(m->node, "query");
+
+  feature = lm_message_node_get_child(ansqry, "feature");
+  for (; feature; feature = feature->next) {
+    const char *v = lm_message_node_get_attribute(feature, "var");
+
+    if (!g_strcmp0(v, NS_CARBONS_2)) {
+      carbons_available();
+    }
+  }
+
+  return LM_HANDLER_RESULT_REMOVE_MESSAGE;
+}
+
 static LmHandlerResult cb_ping(LmMessageHandler *h, LmConnection *c,
                                LmMessage *m, gpointer user_data)
 {
@@ -146,7 +169,8 @@ void xmpp_iq_request(const char *fulljid, const char *xmlns)
 
   iq = lm_message_new_with_sub_type(fulljid, LM_MESSAGE_TYPE_IQ,
                                     LM_MESSAGE_SUB_TYPE_GET);
-  for (i = 0; strcmp(iq_request_handlers[i].xmlns, xmlns) != 0 ; ++i)
+  for (i = 0; iq_request_handlers[i].xmlns &&
+              strcmp(iq_request_handlers[i].xmlns, xmlns) != 0 ; ++i)
        ;
   query = lm_message_node_add_child(iq->node,
                                     iq_request_handlers[i].querytag,
@@ -158,6 +182,11 @@ void xmpp_iq_request(const char *fulljid, const char *xmlns)
     gettimeofday(now, NULL);
     data = (gpointer)now;
     notifier = g_free;
+  } else if (!g_strcmp0(xmlns, NS_DISCO_INFO)) {
+    gchar *servername = get_servername(settings_opt_get("jid"),
+                                       settings_opt_get("server"));
+    lm_message_node_set_attribute(iq->node, "to", servername);
+    g_free(servername);
   }
 
   handler = lm_message_handler_new(iq_request_handlers[i].handler,
@@ -288,7 +317,7 @@ static LmHandlerResult cb_time(LmMessageHandler *h, LmConnection *c,
   }
 
   // Check message contents
-  ansqry = lm_message_node_get_child(m->node, "query");
+  ansqry = lm_message_node_get_child(m->node, "time");
   if (!ansqry) {
     scr_LogPrint(LPRINT_LOGNORM, "Invalid IQ:time result from <%s>!", bjid);
     return LM_HANDLER_RESULT_REMOVE_MESSAGE;
@@ -312,7 +341,7 @@ static LmHandlerResult cb_time(LmMessageHandler *h, LmConnection *c,
                              0, HBB_PREFIX_INFO | HBB_PREFIX_CONT, 0);
     g_free(buf);
   }
-  p = lm_message_node_get_child_value(ansqry, "tz");
+  p = lm_message_node_get_child_value(ansqry, "tzo");
   if (p && *p) {
     buf = g_strdup_printf("TZ:   %s", p);
     scr_WriteIncomingMessage(bjid, buf,
@@ -561,7 +590,7 @@ static LmHandlerResult cb_vcard(LmMessageHandler *h, LmConnection *c,
 static void storage_bookmarks_parse_conference(LmMessageNode *node)
 {
   const char *fjid, *name, *autojoin;
-  const char *pstatus, *awhois;
+  const char *pstatus, *awhois, *fjoins, *group;
   char *bjid;
   GSList *room_elt;
 
@@ -572,13 +601,15 @@ static void storage_bookmarks_parse_conference(LmMessageNode *node)
   autojoin = lm_message_node_get_attribute(node, "autojoin");
   awhois = lm_message_node_get_attribute(node, "autowhois");
   pstatus = lm_message_node_get_child_value(node, "print_status");
+  fjoins = lm_message_node_get_child_value(node, "flag_joins");
+  group = lm_message_node_get_child_value(node, "group");
 
   bjid = jidtodisp(fjid); // Bare jid
 
   // Make sure this is a room (it can be a conversion user->room)
   room_elt = roster_find(bjid, jidsearch, 0);
   if (!room_elt) {
-    room_elt = roster_add_user(bjid, name, NULL, ROSTER_TYPE_ROOM,
+    room_elt = roster_add_user(bjid, name, group, ROSTER_TYPE_ROOM,
                                sub_none, -1);
   } else {
     buddy_settype(room_elt->data, ROSTER_TYPE_ROOM);
@@ -588,6 +619,10 @@ static void storage_bookmarks_parse_conference(LmMessageNode *node)
     // in the roster.
     if (name)
       buddy_setname(room_elt->data, name);
+
+    // The same question for roster group.
+    if (group)
+      buddy_setgroup(room_elt->data, group);
     */
   }
 
@@ -602,18 +637,26 @@ static void storage_bookmarks_parse_conference(LmMessageNode *node)
   }
   if (awhois) {
     enum room_autowhois i = autowhois_default;
-    if (!strcmp(awhois, "1"))
+    if (!strcmp(awhois, "1") || !(strcmp(awhois, "true")))
       i = autowhois_on;
-    else if (!strcmp(awhois, "0"))
+    else if (!strcmp(awhois, "0") || !(strcmp(awhois, "false")))
       i = autowhois_off;
     if (i != autowhois_default)
       buddy_setautowhois(room_elt->data, i);
   }
+  if (fjoins) {
+    enum room_flagjoins i;
+    for (i = flagjoins_none; i <= flagjoins_all; i++)
+      if (!strcasecmp(fjoins, strflagjoins[i]))
+        break;
+    if (i <= flagjoins_all)
+      buddy_setflagjoins(room_elt->data, i);
+  }
 
   // Is autojoin set?
   // If it is, we'll look up for more information (nick? password?) and
   // try to join the room.
-  if (autojoin && !strcmp(autojoin, "1")) {
+  if (autojoin && (!strcmp(autojoin, "1") || !strcmp(autojoin, "true"))) {
     const char *nick, *passwd;
     char *tmpnick = NULL;
     nick = lm_message_node_get_child_value(node, "nick");
@@ -626,6 +669,9 @@ static void storage_bookmarks_parse_conference(LmMessageNode *node)
     g_free(tmpnick);
   }
   g_free(bjid);
+
+  buddylist_build();
+  update_roster = TRUE;
 }
 
 static LmHandlerResult cb_storage_bookmarks(LmMessageHandler *h,
@@ -633,11 +679,13 @@ static LmHandlerResult cb_storage_bookmarks(LmMessageHandler *h,
                                             LmMessage *m, gpointer user_data)
 {
   LmMessageNode *x, *ansqry;
-  char *p;
+  char *p = NULL;
 
   if (lm_message_get_sub_type(m) == LM_MESSAGE_SUB_TYPE_ERROR) {
+    LmMessageNode *error = lm_message_node_get_child(m->node, "error");
     // No server support, or no bookmarks?
-    p = m->node->children->name;
+    if (error && error->children)
+      p = error->children->name;
     if (p && !strcmp(p, "item-not-found")) {
       // item-no-found means the server has Private Storage, but it's
       // currently empty.
@@ -646,9 +694,9 @@ static LmHandlerResult cb_storage_bookmarks(LmMessageHandler *h,
       bookmarks = lm_message_node_new("storage", "storage:bookmarks");
       // We return 0 so that the IQ error message be
       // not displayed, as it isn't a real error.
-      return LM_HANDLER_RESULT_REMOVE_MESSAGE;
-    }
-    return LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS; // Unhandled error
+    } else
+      scr_LogPrint(LPRINT_LOGNORM, "Server does not support bookmarks storage.");
+    return LM_HANDLER_RESULT_REMOVE_MESSAGE;
   }
 
   ansqry = lm_message_node_get_child(m->node, "query");
@@ -680,9 +728,11 @@ static LmHandlerResult cb_storage_rosternotes(LmMessageHandler *h,
   LmMessageNode *ansqry;
 
   if (lm_message_get_sub_type(m) == LM_MESSAGE_SUB_TYPE_ERROR) {
-    const char *p;
+    const char *p = NULL;
+    LmMessageNode *error = lm_message_node_get_child(m->node, "error");
     // No server support, or no roster notes?
-    p = m->node->children->name;
+    if (error && error->children)
+      p = error->children->name;
     if (p && !strcmp(p, "item-not-found")) {
       // item-no-found means the server has Private Storage, but it's
       // currently empty.
@@ -691,9 +741,9 @@ static LmHandlerResult cb_storage_rosternotes(LmMessageHandler *h,
       rosternotes = lm_message_node_new("storage", "storage:rosternotes");
       // We return 0 so that the IQ error message be
       // not displayed, as it isn't a real error.
-      return LM_HANDLER_RESULT_REMOVE_MESSAGE;
-    }
-    return LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS; // Unhandled error
+    } else
+      scr_LogPrint(LPRINT_LOGNORM, "Server does not support roster notes storage.");
+    return LM_HANDLER_RESULT_REMOVE_MESSAGE;
   }
 
   ansqry = lm_message_node_get_child(m->node, "query");
